import gleam/dynamic/decode
import gleam/json
import gleam/option.{type Option}
import spacetraders_sdk/endpoints/fleet
import spacetraders_sdk/models/agent.{type Agent}
import spacetraders_sdk/models/contract.{type Contract}
import spacetraders_sdk/models/contract_id.{type ContractId}
import spacetraders_sdk/models/ship_cargo.{type ShipCargo}
import spacetraders_sdk/models/ship_symbol.{type ShipSymbol}
import spacetraders_sdk/models/trade_symbol.{type TradeSymbol}
import spacetraders_sdk/utils/api.{type ApiResponse, type PagedData}
import spacetraders_sdk/utils/auth.{type AgentToken, AgentAuth}

pub fn list_contracts(
  token: AgentToken,
  page: Option(Int),
  limit: Option(Int),
) -> ApiResponse(PagedData(List(Contract))) {
  let request = api.get_page(AgentAuth(token), "/my/contracts", page, limit)
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_paged_data_response(response, decode.list(contract.decoder()))
    _ -> api.parse_error_response(response)
  }
}

pub fn get_contract(
  token: AgentToken,
  contract_id: ContractId,
) -> ApiResponse(Contract) {
  let request =
    api.get(
      AgentAuth(token),
      "/my/contracts/" <> contract_id.to_string(contract_id),
    )
  use response <- api.try_send(request)
  case response.status {
    200 -> api.parse_data_response(response, contract.decoder())
    _ -> api.parse_error_response(response)
  }
}

pub type ContractAccepted {
  ContractAccepted(contract: Contract, agent: Agent)
}

pub fn accept_contract(
  token: AgentToken,
  contract_id: ContractId,
) -> ApiResponse(ContractAccepted) {
  let request =
    api.post(
      AgentAuth(token),
      "/my/contracts/" <> contract_id.to_string(contract_id) <> "/accept",
    )
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_data_response(response, {
        use contract <- decode.field("contract", contract.decoder())
        use agent <- decode.field("agent", agent.decoder())
        decode.success(ContractAccepted(contract:, agent:))
      })
    _ -> api.parse_error_response(response)
  }
}

pub type ContractFulfilled {
  ContractFulfilled(contract: Contract, agent: Agent)
}

pub fn fulfill_contract(
  token: AgentToken,
  contract_id: ContractId,
) -> ApiResponse(ContractFulfilled) {
  let request =
    api.post(
      AgentAuth(token),
      "/my/contracts/" <> contract_id.to_string(contract_id) <> "/fulfill",
    )
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_data_response(response, {
        use contract <- decode.field("contract", contract.decoder())
        use agent <- decode.field("agent", agent.decoder())
        decode.success(ContractFulfilled(contract:, agent:))
      })
    _ -> api.parse_error_response(response)
  }
}

pub type ContractCargoDelivered {
  ContractCargoDelivered(contract: Contract, cargo: ShipCargo)
}

pub fn deliver_contract_cargo(
  token: AgentToken,
  contract_id: ContractId,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> ApiResponse(ContractCargoDelivered) {
  let request =
    api.post_json(
      AgentAuth(token),
      "/my/contracts/" <> contract_id.to_string(contract_id) <> "/deliver",
      json.object([
        #("shipSymbol", ship_symbol.encode(ship_symbol)),
        #("tradeSymbol", trade_symbol.encode(trade_symbol)),
        #("units", json.int(units)),
      ]),
    )
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_data_response(response, {
        use contract <- decode.field("contract", contract.decoder())
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        decode.success(ContractCargoDelivered(contract:, cargo:))
      })
    _ -> api.parse_error_response(response)
  }
}

pub const negotiate_contract = fleet.negotiate_contract
